from .base import WeDeliverCore
from .app_decorators.app_entry import route
from .helpers.log_config import init_logger
from .helpers.config import Config
from .helpers.kafka_producer import Producer
from .helpers.topics import Topics
from .helpers.micro_fetcher import MicroFetcher

__all__ = [
    "WeDeliverCore",
    "route",
    "Config",
    "Producer",
    "init_logger",
    "Topics",
    "MicroFetcher",
]



import json
from functools import wraps
from sqlalchemy.orm.base import object_mapper
import flask_sqlalchemy
from flask import request
from marshmallow import ValidationError
from sqlalchemy.orm.exc import UnmappedInstanceError

from wedeliver_core.helpers.exceptions import AppValidationError


def is_mapped(obj):
    try:
        data = obj
        if isinstance(data, list) and len(data):
            data = obj[0]
        object_mapper(data)
    except UnmappedInstanceError:
        return False
    return True


def serializer(schema=None, many=False):
    def factory(func):
        @wraps(func)
        def decorator(*args, **kwargs):
            if not schema:
                return func(*args, **kwargs)

            appended_kws = kwargs.pop('appended_kws', None)

            try:
                client_data = dict()

                if request.json:
                    client_data = request.json
                elif request.form:
                    client_data = request.form.to_dict()
                    for key, value in client_data.items():
                        try:
                            # look for strings like objects to convert them to python objects
                            client_data[key] = json.loads(value)
                        except json.JSONDecodeError as e:
                            pass

                if request.args:
                    client_data.update(request.args.to_dict())

                inputs = client_data  # .to_dict()
                if appended_kws:
                    inputs.update(appended_kws)
                result = schema(many=many).load(inputs)
            except ValidationError as e:
                raise AppValidationError(e.messages)

            kwargs.update(dict(validated_data=result))

            # if schema and request.method == "GET":
            try:
                result = func(*args, **kwargs)
                if isinstance(result, flask_sqlalchemy.Pagination):
                    items = schema(many=isinstance(result.items, list)).dump(result.items)
                    output = dict(
                        items=items,
                        total=result.total,
                        next_num=result.next_num,
                        prev_num=result.prev_num,
                        page=result.page,
                        per_page=result.per_page
                    )
                elif is_mapped(result):  # is model instance
                    output = schema(many=isinstance(result, list)).dump(result)
                else:
                    output = result
            except ValidationError as e:
                raise AppValidationError(e.messages)

            return output

            # return func(*args, **kwargs)

        return decorator

    return factory



from wedeliver_core import WeDeliverCore
from wedeliver_core.helpers.sql import sql
import importlib
import os


def fetch_relational_data(fields,
                          table_name,
                          column_name,
                          compair_operator,
                          column_values, functions=None):
    app = WeDeliverCore.getApp()
    db = app.extensions['sqlalchemy'].db

    relational_data_result = []
    if table_name:
        query = """
        SELECT {fields}
        FROM {table_name}
        WHERE {column_name} {compair_operator} {column_values}
        """.format(
            fields=', '.join(fields),
            table_name=table_name,
            column_name=column_name,
            compair_operator=compair_operator,
            column_values='({})'.format(
                ', '.join([str(val) for val in column_values])) if compair_operator == 'IN' else column_values,
        )
        relational_data_result = sql(query, db)

    def _update_relational_data_result(function_result, fields):
        if relational_data_result:
            for rd in relational_data_result:
                for f in fields:
                    rd[f] = function_result.get(f) if isinstance(function_result,
                                                                 dict) else function_result.__dict__.get(f)
        else:
            fields_dict = dict()
            for f in fields:
                fields_dict[f] = function_result.get(f) if isinstance(function_result,
                                                                      dict) else function_result.__dict__.get(f)
            relational_data_result.append(fields_dict)

    if functions and isinstance(functions, list):
        for func in functions:
            function_file, function_call = os.path.splitext(func.get('name'))
            m = importlib.import_module(function_file)
            method = getattr(m, function_call[1:])

            function_result = method(**func.get('params'))

            fields = func.get('fields')
            if function_result:
                if isinstance(function_result, list):
                    for row in function_result:
                        _update_relational_data_result(row, fields)
                else:
                    _update_relational_data_result(function_result, fields)

    return relational_data_result



class WeDeliverCore:
    __app = None

    @staticmethod
    def getApp():
        """ Static access method. """
        if WeDeliverCore.__app == None:
            WeDeliverCore()
        return WeDeliverCore.__app

    def __init__(self, app=None):
        """ Virtually private constructor. """
        if WeDeliverCore.__app != None:
            raise Exception("This class is a singleton!")
        else:
            WeDeliverCore.__app = app
